#!/usr/bin/env python3

# Open SSH connections through AWS Session Manager
#
# See https://aws.nz/aws-utils/ssm-ssh for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list available instances, resolve instance names,
# and host names, etc. In the end it executes 'ssh' with the correct
# parameters to actually start the SSH session.

import os
import sys
import logging
import argparse
import botocore.exceptions

from .common import *
from .resolver import InstanceResolver

logger = logging.getLogger()

def parse_args(argv):
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser, long_only=True)

    group_instance = parser.add_argument_group('Instance Selection')
    group_instance.add_argument('--list', dest='list', action="store_true", help='List instances available for SSM Session')

    parser.description = 'Open SSH connection through Session Manager'
    parser.epilog = f'''
IMPORTANT: instances must be registered in AWS Systems Manager (SSM)
before you can start a shell session! Instances not registered in SSM
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ssm-ssh for more info and usage examples.

Author: Michael Ludvig
'''

    # Parse supplied arguments
    args, extra_args = parser.parse_known_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of INSTANCE or --list
    if bool(extra_args) + bool(args.list) != 1:
        parser.error("Specify either --list or SSH Options including instance name")

    return args, extra_args

def start_ssh_session(ssh_args, profile=None, region=None):
    aws_args = ""
    if profile:
        aws_args += f"--profile {profile} "
    if region:
        aws_args += f"--region {region} "
    proxy_option = f"-o ProxyCommand='aws {aws_args} ssm start-session --target %h --document-name AWS-StartSSHSession --parameters portNumber=%p'"
    ssh_args = " ".join([ f"'{arg}'" for arg in ssh_args ])
    command = f'ssh {proxy_option} {ssh_args}'
    logger.info("Running: %s", command)
    os.system(command)

def main():
    global logger

    ## Split command line to main args and optional command to run
    args, extra_args = parse_args(sys.argv[1:])

    if args.log_level == logging.INFO:
        extra_args.append('-v')
    if args.log_level == logging.DEBUG:
        extra_args.append('-vv')

    logger = configure_logging("ssm-ssh", args.log_level)

    if not verify_plugin_version("1.1.23", logger):
        quit(1)

    try:
        instance_resolver = InstanceResolver(args)

        if args.list:
            instance_resolver.print_list()
            quit(0)

        # Loop through all possible instance names and try to resolve them.
        ssh_args = []
        instance_id = None
        for arg in extra_args:
            # If we already have instance id just copy the args
            if instance_id:
                ssh_args.append(arg)
                continue

            # Some args that can't be an instance name
            if arg.startswith('-') or arg.find(':') > -1 or arg.find(os.path.sep) > -1:
                ssh_args.append(arg)
                continue

            # This may be an instance name - try to resolve it
            login_name = None
            if arg.find('@') > -1:  # username@hostname format
                login_name, instance = arg.split('@', 1)
            else:
                instance = arg

            instance_id = instance_resolver.resolve_instance(instance)
            if not instance_id:
                # Not resolved as an instance name - put back to args
                ssh_args.append(arg)
                continue

            # Woohoo we've got an instance id!
            logger.info(f"Resolved instance name '{instance}' to '{instance_id}'")
            ssh_args.append(instance_id)

            if login_name:
                ssh_args.extend(['-l', login_name])

        if not instance_id:
            logger.warning("Could not resolve Instance ID for '%s'", instance)
            logger.warning("Perhaps the '%s' is not registered in SSM?", instance)
            quit(1)

        start_ssh_session(ssh_args=ssh_args, profile=args.profile, region=args.region)

    except (botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError) as e:
        logger.error(e)
        quit(1)

if __name__ == "__main__":
    main()



import sys
import logging
import subprocess
import boto3
import packaging.version
from . import __version__ as ssm_tools_version

__all__ = []

# ---------------------------------------------------------

__all__.append("configure_logging")
def configure_logging(logger_name, level):
    """
    Configure logging format and level.
    """
    streamHandler = logging.StreamHandler()
    formatter = logging.Formatter(
        "[%(name)s] %(levelname)s: %(message)s"
    )
    streamHandler.setFormatter(formatter)
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    logger.addHandler(streamHandler)
    logger.debug("Logging level set to DEBUG")

    return logger

# ---------------------------------------------------------

__all__.append("add_general_parameters")
def add_general_parameters(parser, long_only = False):
    """
    Add General Options used by all ssm-* tools.
    """

    # Remove short options if long_only==True
    def _get_opts(opt_long, opt_short):
        opts = [ opt_long ]
        if not long_only:
            opts.append(opt_short)
        return opts

    group_general = parser.add_argument_group('General Options')
    group_general.add_argument(*_get_opts('--profile', '-p'), dest='profile', type=str, help='Configuration profile from ~/.aws/{credentials,config}')
    group_general.add_argument(*_get_opts('--region',  '-g'), dest='region', type=str, help='Set / override AWS region.')
    group_general.add_argument(*_get_opts('--verbose', '-v'), action='store_const', dest='log_level', const=logging.INFO, default=logging.WARN, help='Increase log_level level')
    group_general.add_argument(*_get_opts('--debug',   '-d'), action='store_const', dest='log_level', const=logging.DEBUG, help='Increase log_level level')
    group_general.add_argument(*_get_opts('--version', '-V'), action='store_true', dest='show_version', help=f'Show package version and exit. Version is {ssm_tools_version}')
    group_general.add_argument(*_get_opts('--help',    '-h'), action="help", help='Print this help and exit')

    return group_general

# ---------------------------------------------------------

__all__.append("show_version")
def show_version(args):
    """
    Show package version and exit.
    """
    version_string = f"ssm-tools/{ssm_tools_version}"
    if args.log_level <= logging.INFO:
        version_string += f" python/{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        version_string += f" boto3/{boto3.__version__}"
    print(version_string)
    sys.exit(0)

# ---------------------------------------------------------

__all__.append("bytes_to_human")
def bytes_to_human(size):
    """
    Convert Bytes to more readable units
    """
    units = [ 'B', 'kB', 'MB', 'GB', 'TB' ]
    unit_idx = 0    # Start with Bytes
    while unit_idx < len(units)-1:
        if size < 2048:
            break
        size /= 1024.0
        unit_idx += 1
    return size, units[unit_idx]

# ---------------------------------------------------------

__all__.append("seconds_to_human")
def seconds_to_human(seconds, decimal=3):
    """
    Convert seconds to HH:MM:SS[.SSS]

    If decimal==0 only full seconds are used.
    """
    secs = int(seconds)
    fraction = seconds - secs

    mins = int(secs / 60)
    secs = secs % 60

    hours = int(mins / 60)
    mins = mins % 60

    ret = f"{hours:02d}:{mins:02d}:{secs:02d}"
    if decimal:
        fraction = int(fraction * (10**decimal))
        ret += f".{fraction:0{decimal}d}"

    return ret

# ---------------------------------------------------------

__all__.append("verify_plugin_version")
def verify_plugin_version(version_required, logger):
    """
    Verify that a session-manager-plugin is installed
    and is of a required version or newer.
    """
    session_manager_plugin = 'session-manager-plugin'

    try:
        result = subprocess.run([session_manager_plugin, '--version'], stdout=subprocess.PIPE)
        plugin_version = result.stdout.decode('ascii').strip()
        logger.debug(f"{session_manager_plugin} version {plugin_version}")

        if packaging.version.parse(plugin_version) >= packaging.version.parse(version_required):
            return True

        logger.error(f"ERROR: session-manager-plugin version {plugin_version} is installed, {version_required} is required")
    except FileNotFoundError as e:
        logger.error(f"ERROR: {session_manager_plugin} not installed")

    logger.error("ERROR: Check out https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html for instructions")

    return False



#!/usr/bin/env python3

# Convenience wrapper around 'aws ecs execute-command'
#
# See https://aws.nz/aws-utils/ecs-session for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list the available containers across all your ECS clusters.
# In the end it executes 'aws ecs execute-command' with the appropriate parameters.
# Supports both EC2 and Fargate ECS tasks.

import os
import sys
import logging
import signal
import argparse
import botocore.exceptions

from .common import *
from .resolver import ContainerResolver

logger = logging.getLogger()

def parse_args(argv):
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser)

    group_container = parser.add_argument_group('Container Selection')
    group_container.add_argument('CONTAINER', nargs='?', help='Task ID, Container Name or IP address')
    group_container.add_argument('--list', '-l', dest='list', action="store_true", help='List available containers configured for ECS RunTask')

    group_session = parser.add_argument_group('Session Parameters')
    group_session.add_argument('--command', dest='command', metavar='COMMAND', default='/bin/sh', help='Command to run inside the container. Default: /bin/sh')

    parser.description = 'Execute "ECS Run Task" on a given container'
    parser.epilog = f'''
IMPORTANT: containers must have "execute-command" setting enabled or they
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ecs-session for more info and usage examples.

Author: Michael Ludvig
'''

    # Parse supplied arguments
    args, extras = parser.parse_known_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of CONTAINER or --list
    if bool(args.CONTAINER) + bool(args.list) != 1:
        parser.error("Specify either CONTAINER or --list")

    return args, extras

def start_session(container, args, command):
    aws_args = ""
    if args.profile:
        aws_args += f"--profile {args.profile} "
    if args.region:
        aws_args += f"--region {args.region} "

    command = (
        f'aws {aws_args} ecs execute-command '
            f'--cluster {container["cluster_arn"]} '
            f'--task {container["task_arn"]} '
            f'--container {container["container_name"]} '
            f'--command \'{command}\' --interactive '
    )
    logger.info("Running: %s", command)
    os.system(command)

def main():
    ## Split command line to main args and optional command to run
    args, extras = parse_args(sys.argv[1:])

    global logger
    logger = configure_logging("ecs-session", args.log_level)

    try:
        if args.list:
            ContainerResolver(args).print_list()
            quit(0)

        container = ContainerResolver(args).resolve_container(args.CONTAINER)

        if not container:
            logger.warning("Could not find any container matching '%s'", args.CONTAINER)
            quit(1)

        start_session(container, args, args.command)

    except (botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError) as e:
        logger.error(e)
        quit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import re
import logging
import botocore.credentials
import botocore.session
import boto3

logger = logging.getLogger("ssm-session")

class CommonResolver():
    def __init__(self, args):
        # aws-cli compatible MFA cache
        cli_cache = os.path.join(os.path.expanduser('~'),'.aws/cli/cache')

        # Construct boto3 session with MFA cache
        self.session = boto3.session.Session(profile_name=args.profile, region_name=args.region)
        self.session._session.get_component('credential_provider').get_provider('assume-role').cache = botocore.credentials.JSONFileCache(cli_cache)


class InstanceResolver(CommonResolver):
    def __init__(self, args):
        super().__init__(args)

        # Create boto3 clients from session
        self.ssm_client = self.session.client('ssm')
        self.ec2_client = self.session.client('ec2')

    def get_list(self):
        def _try_append(_list, _dict, _key):
            if _key in _dict:
                _list.append(_dict[_key])

        items = {}

        # List instances from SSM
        logger.debug("Fetching SSM inventory")
        paginator = self.ssm_client.get_paginator('get_inventory')
        response_iterator = paginator.paginate()
        for inventory in response_iterator:
            for entity in inventory["Entities"]:
                try:
                    content = entity['Data']['AWS:InstanceInformation']["Content"][0]

                    # At the moment we only support EC2 Instances and ManagedInstances
                    if content["ResourceType"] not in [ "EC2Instance", "ManagedInstance" ]:
                        logger.warning("Unknown instance type: %s: %s", entity['Id'], content['ResourceType'])
                        logger.debug(entity)
                        continue

                    # Ignore Terminated instances
                    if content.get("InstanceStatus") == "Terminated":
                        logger.debug("Ignoring terminated instance: %s", entity)
                        continue

                    # Add to the list
                    instance_id = content['InstanceId']
                    items[instance_id] = {
                        "InstanceId": instance_id,
                        "InstanceName": "",
                        "HostName": content.get("ComputerName", ""),
                        "Addresses": [ content.get("IpAddress") ],
                    }
                    logger.debug("Added instance: %s: %r", instance_id, items[instance_id])
                except (KeyError, ValueError):
                    logger.warning("SSM inventory entity not recognised: %s", entity)
                    continue

        # Add attributes from EC2
        paginator = self.ec2_client.get_paginator('describe_instances')
        ec2_instance_ids = list(filter(lambda x: x.startswith("i-"), items))

        tries = 5
        while tries:
            # The SSM inventory sometimes returns instances that have been terminated
            # a short while ago which makes the following call fail
            # with InvalidInstanceID.NotFound exception. We'll try and remove the invalid
            # instance ids a {tries} times or until we succeed. If unsuccessful we'll remove
            # the list obtained from SSM without extra details (host name, public IPs, etc).
            # This mostly / only affects accounts with high churn of starting / stopping
            # instances - most users will pass this loop only once.
            try:
                response_iterator = paginator.paginate(InstanceIds=ec2_instance_ids)
                for reservations in response_iterator:
                    for reservation in reservations['Reservations']:
                        for instance in reservation['Instances']:
                            instance_id = instance['InstanceId']
                            if not instance_id in items:
                                continue

                            # Find instance IPs
                            items[instance_id]['Addresses'] = []
                            _try_append(items[instance_id]['Addresses'], instance, 'PrivateIpAddress')
                            _try_append(items[instance_id]['Addresses'], instance, 'PublicIpAddress')

                            # Find instance name from tag Name
                            for tag in instance.get('Tags', []):
                                if tag['Key'] == 'Name' and tag['Value']:
                                    items[instance_id]['InstanceName'] = tag['Value']

                            logger.debug("Updated instance: %s: %r", instance_id, items[instance_id])
                    return items

            except botocore.exceptions.ClientError as ex:
                if ex.response.get('Error', {}).get('Code', '') != 'InvalidInstanceID.NotFound':
                    raise
                message = ex.response.get('Error', {}).get('Message', '')
                if not message.startswith("The instance ID") or not message.endswith("not exist"):
                    logger.warning("Unexpected InvalidInstanceID.NotFound message:", message)
                # Try to extract instace ids ...
                remove_instance_ids = re.findall('i-[0-9a-f]+', message)
                logger.debug("Removing non-existent InstanceIds: %s", remove_instance_ids)
                # Remove the failed ids from the list and try again
                ec2_instance_ids = list(set(ec2_instance_ids) - set(remove_instance_ids))
                tries -= 1

        if not tries:
            logger.warning("Unable to list instance details. Some instance names and IPs may be missing.")

        return items

    def print_list(self):
        hostname_len = 1    # Minimum of 1 char, otherwise f-string below fails for empty hostnames
        instname_len = 1

        items = self.get_list().values()

        if not items:
            logger.warning("No instances registered in SSM!")
            return

        items = list(items)
        items.sort(key=lambda x: x.get('InstanceName') or x.get('HostName'))

        for item in items:
            hostname_len = max(hostname_len, len(item['HostName']))
            instname_len = max(instname_len, len(item['InstanceName']))

        for item in items:
            print(f"{item['InstanceId']:20}   {item['HostName']:{hostname_len}}   {item['InstanceName']:{instname_len}}   {' '.join(item['Addresses'])}")

    def resolve_instance(self, instance):
        # Is it a valid Instance ID?
        if re.match('^i-[a-f0-9]+$', instance):
            return instance

        # It is not - find it in the list
        instances = []

        items = self.get_list()
        for instance_id in items:
            item = items[instance_id]
            if instance.lower() in [item['HostName'].lower(), item['InstanceName'].lower()] + item['Addresses']:
                instances.append(instance_id)

        if not instances:
            return None

        if len(instances) > 1:
            logger.warning("Found %d instances for '%s': %s", len(instances), instance, " ".join(instances))
            logger.warning("Use INSTANCE_ID to connect to a specific one")
            quit(1)

        # Found only one instance - return it
        return instances[0]

class ContainerResolver(CommonResolver):
    def __init__(self, args):
        super().__init__(args)

        # Create boto3 clients from session
        self.ecs_client = self.session.client('ecs')

        self.containers = []
        self._tasks = {}

    def add_container(self, container):
        _task_parsed = container['taskArn'].split(":")[-1].split("/")
        self.containers.append({
            "cluster_name": _task_parsed[1],
            "task_id": _task_parsed[2],
            "cluster_arn": self._tasks[container['taskArn']]['clusterArn'],
            "task_arn": container['taskArn'],
            "group_name": self._tasks[container['taskArn']]['group'],
            "container_name": container['name'],
            "container_ip": container['networkInterfaces'][0]['privateIpv4Address'],
        })

    def get_list(self):
        def _try_append(_list, _dict, _key):
            if _key in _dict:
                _list.append(_dict[_key])

        items = {}

        # List ECS Clusters
        clusters = []
        logger.debug("Listing ECS Clusters")
        paginator = self.ecs_client.get_paginator('list_clusters')
        for page in paginator.paginate():
            clusters.extend(page['clusterArns'])

        if not clusters:
            logger.warning("No ECS Clusters found.")
            return []

        # List tasks in each cluster
        paginator = self.ecs_client.get_paginator('list_tasks')
        for cluster in clusters:
            logger.debug("Listing tasks in cluster: %s", cluster)

            # maxResults must be <= 100 because describe_tasks() doesn't accept more than that
            for page in paginator.paginate(cluster=cluster, maxResults=100):
                response = self.ecs_client.describe_tasks(cluster=cluster, tasks=page['taskArns'])

                # Filter containers that have a running ExecuteCommandAgent
                for task in response['tasks']:
                    self._tasks[task['taskArn']] = task
                    for container in task['containers']:
                        if not 'managedAgents' in container:
                            continue
                        for agent in container['managedAgents']:
                            if agent['name'] == 'ExecuteCommandAgent' and agent['lastStatus'] == 'RUNNING':
                                self.add_container(container)

        return self.containers

    def print_containers(self, containers):
        max_len = {}
        for container in containers:
            for key in container.keys():
                if not key in max_len:
                    max_len[key] = len(container[key])
                else:
                    max_len[key] = max(max_len[key], len(container[key]))
        containers.sort(key = lambda x: [x['cluster_name'], x['container_name']])
        for container in containers:
            print(f"{container['cluster_name']:{max_len['cluster_name']}}  {container['group_name']:{max_len['group_name']}}  {container['task_id']:{max_len['task_id']}}  {container['container_name']:{max_len['container_name']}}  {container['container_ip']:{max_len['container_ip']}}")

    def print_list(self):
        containers = self.get_list()

        if not containers:
            logger.warning("No Execute-Command capable contaianers found!")
            quit(1)

        self.print_containers(containers)

    def resolve_container(self, keyword):
        containers = self.get_list()

        if not containers:
            logger.warning("No Execute-Command capable contaianers found!")
            quit(1)

        logger.debug("Searching for: %s", keyword)

        candidates = []
        for container in containers:
            if keyword in (container['group_name'], container['task_id'], container['container_name'], container['container_ip']):
                candidates.append(container)
        if not candidates:
            logger.warning("No container found for: %s", keyword)
            quit(1)
        elif len(candidates) == 1:
            self.print_containers(candidates)
            return candidates[0]
        else:
            logger.warning("Found %d instances for: %s", len(containers), keyword)
            logger.warning("Use Container IP or Task ID to connect to a specific one")
            self.print_containers(candidates)
            quit(1)



#!/usr/bin/env python3

# Set up IP tunnel through SSM-enabled instance.
#
# See https://aws.nz/aws-utils/ssm-tunnel for more info.
#
# Author: Michael Ludvig (https://aws.nz)

import os
import sys
import time
import copy
import errno
import threading
import random
import struct
import select
import fcntl
import argparse
import ipaddress
from base64 import b64encode, b64decode

import pexpect
import botocore.exceptions

from .common import *
from .talker import SsmTalker
from .resolver import InstanceResolver

logger_name = "ssm-tunnel"
tunnel_cidr = "100.64.0.0/16"
keepalive_sec = 10

def parse_args():
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    group_general = add_general_parameters(parser)

    group_instance = parser.add_argument_group('Instance Selection')
    group_instance.add_argument('INSTANCE', nargs='?', help='Instance ID, Name, Host name or IP address')
    group_instance.add_argument('--list', '-l', dest='list', action="store_true", help='List instances registered in SSM.')

    group_network = parser.add_argument_group('Networking Options')
    group_network.add_argument('--route', '-r', metavar="ROUTE", dest="routes", type=str, action="append",
        help='CIDR(s) to route through this tunnel. May be used multiple times.')
    group_network.add_argument('--tunnel-cidr', metavar="CIDR", type=str, default=tunnel_cidr, help=f'''By default
        the tunnel endpoint IPs are randomly assigned from the reserved {tunnel_cidr} block (RFC6598).
        This should be ok for most users.''')
    group_network.add_argument('--up-down', metavar="SCRIPT", dest='updown_script', type=str, help='''Script to call
        during tunnel start up and close down. Check out 'ssm-tunnel-updown.dns-example' that
        supports setting a custom DNS server when the tunnel goes up.''')

    parser.description = 'Start IP tunnel to a given SSM instance'
    parser.epilog = f'''
IMPORTANT: instances must be registered in AWS Systems Manager (SSM)
before you can copy files to/from them! Instances not registered in SSM
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ssm-tunnel for more info and usage examples.

Author: Michael Ludvig
'''

    # Parse supplied arguments
    args = parser.parse_args()

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of INSTANCE or --list
    if bool(args.INSTANCE) + bool(args.list) != 1:
        parser.error("Specify either INSTANCE or --list")

    return args

class SsmTunnel(SsmTalker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Stats structure
        self.stats = { 'ts': 0, 'l2r': 0, 'r2l': 0 }
        self.stats_lock = threading.Lock()
        self.stats_secs = 10
        self.stats_refresh = 0.5        # Print stats every this many seconds

        self._exiting = False

        self.tun_name = self._tun_fd = None
        self.local_ip = self.remote_ip = self.routes = None
        self.updown_script = self.updown_up_success = None

    def run_command(self, command, assert_0=True):
        self._logger.debug("command: %s", command)
        ret = os.system(command)
        if assert_0:
            assert ret == 0

    def open_remote_tunnel(self):
        self._logger.debug('Creating tunnel')

        # Open remote tun0 device
        self._child.sendline(f"ssm-tunnel-agent {self.remote_ip} {self.local_ip}")
        patterns = ['# Agent device .* is ready', 'command not found']
        match = self._child.expect(patterns)
        if match != 0:  # Index matched in the 'patterns'
            self._logger.error("Unable to establish the tunnel!")
            self._logger.error("ssm-tunnel-agent: command not found on the target instance %s.", self._instance_id)
            self._logger.error("Use 'ssm-session %s' and then run 'sudo pip install aws-ssm-tunnel-agent' to install it.", self._instance_id)
            quit(1)
        self._logger.debug(self._child.after)

    def open_local_tunnel(self):
        tun_suffix = ".".join(self.local_ip.split(".")[2:])
        self.tun_name = f"tunSSM.{tun_suffix}"

        self.create_tun()
        self._tun_fd = self.open_tun()

        self._logger.debug(f"# Local device {self.tun_name} is ready")
        self._logger.info(f"Local IP: {self.local_ip} / Remote IP: {self.remote_ip}")

    def create_tun(self):
        try:
            user_id = os.getuid()
            self.run_command(f"sudo ip tuntap add {self.tun_name} mode tun user {user_id}")
            self.run_command(f"sudo ip addr add {self.local_ip} peer {self.remote_ip} dev {self.tun_name}")
            self.run_command(f"sudo ip link set {self.tun_name} up")
            # Configure routes
            for route in self.routes:
                self.run_command(f"sudo ip route add {route} via {self.remote_ip}")
        except AssertionError:
            self.delete_tun()
            quit(1)
        except Exception as e:
            self._logger.exception(e)
            self.delete_tun()
            raise

    def delete_tun(self):
        # We don't check return code here - best effort to close and delete the device
        if self._tun_fd is not None:
            try:
                os.close(self._tun_fd)
                self._tun_fd = None
            except Exception as e:
                self._logger.exception(e)
        if self.tun_name is not None:
            self.run_command(f"sudo ip link set {self.tun_name} down", assert_0=False)
            self.run_command(f"sudo ip tuntap del {self.tun_name} mode tun", assert_0=False)
            self.tun_name = None

    def open_tun(self):
        TUNSETIFF = 0x400454ca
        IFF_TUN = 0x0001

        tun_fd = os.open("/dev/net/tun", os.O_RDWR)

        flags = IFF_TUN
        ifr = struct.pack('16sH22s', self.tun_name.encode(), flags, b'\x00'*22)
        fcntl.ioctl(tun_fd, TUNSETIFF, ifr)

        return tun_fd

    def local_to_remote(self):
        last_ts = time.time()
        while True:
            if self._exiting:
                break
            try:
                r, w, x = select.select([self._tun_fd], [], [], 1)
                if not self._tun_fd in r:
                    if last_ts + keepalive_sec < time.time():
                        # Keepalive timeout - send '#'
                        self._child.sendline("#")
                        last_ts = time.time()
                    continue
                buf = os.read(self._tun_fd, 1504)     # Virtual GRE header adds 4 bytes
                self._child.sendline("%{}".format(b64encode(buf).decode('ascii')))
            except OSError as e:
                if e.errno == errno.EBADF and self._exiting:
                    break

            last_ts = time.time()
            # Update stats
            self.stats_lock.acquire()
            self.stats['l2r'] += len(buf)
            self.stats_lock.release()

        self._logger.debug("local_to_remote() has exited.")

    def remote_to_local(self):
        while True:
            if self._exiting:
                break
            try:
                line = self._child.readline()
            except pexpect.exceptions.TIMEOUT:
                # This is a long timeout, 30 sec, not very useful
                continue
            if type(self._child.after) == pexpect.exceptions.EOF:
                self._logger.warn("Received unexpected EOF - tunnel went down?")
                self._exiting = True
                break
            if not line or line[0] != '%':
                continue

            buf = b64decode(line[1:].strip('\r\n'))
            os.write(self._tun_fd, buf)
            # Update stats
            self.stats_lock.acquire()
            self.stats['r2l'] += len(buf)
            self.stats_lock.release()

        self._logger.debug("remote_to_local() has exited.")

    def process_traffic(self):
        tr_l2r = threading.Thread(target=self.local_to_remote, args=[])
        tr_l2r.daemon = True
        tr_l2r.start()

        tr_r2l = threading.Thread(target=self.remote_to_local, args=[])
        tr_r2l.daemon = True
        tr_r2l.start()

        try:
            self.display_stats()

        except KeyboardInterrupt:
            print("")   # Just to avoid "^C" at the end of line

    def run_updown(self, status):
        if not self.updown_script:
            return

        if status == "down" and not self.updown_up_success:
            # If 'up' failed we are immediately called with 'down' - don't do anything.
            return

        routes = " ".join(self.routes)
        try:
            cmd = f"{self.updown_script} {status} {self.tun_name} {self.local_ip} {self.remote_ip} {routes}"
            self._logger.info(f"Running --up-down script: {cmd}")
            self.run_command(cmd)
            self.updown_up_success = True
        except AssertionError:
            self._logger.error(f'Updown script {self.updown_script} exitted with error.')
            sys.exit(1)

    def start(self, local_ip, remote_ip, routes, updown_script):
        self.local_ip = local_ip
        self.remote_ip = remote_ip
        self.routes = routes
        self.updown_script = updown_script

        try:
            self.open_remote_tunnel()
            self.open_local_tunnel()
            self.run_updown("up")
            self.process_traffic()

        finally:
            self._logger.info('Closing tunnel, please wait...')
            self.run_updown("down")
            self.exit()
            self._exiting = True
            self.delete_tun()


    def display_stats(self):
        def _erase_line():
            print('\r\x1B[K', end="")   # Erase line

        stat_history = [self.stats]
        stat_history_len = int(self.stats_secs / self.stats_refresh)
        start_ts = time.time()

        while True:
            time.sleep(self.stats_refresh)

            # Take another 'stat' snapshot
            self.stats_lock.acquire()
            stat_history.insert(1, copy.copy(self.stats))
            self.stats_lock.release()
            stat_history[1]['ts'] = time.time()

            # Calculate sliding window average
            if stat_history[1]['ts'] > stat_history[-1]['ts']:
                l2r_avg = (stat_history[1]['l2r'] - stat_history[-1]['l2r'])/(stat_history[1]['ts'] - stat_history[-1]['ts'])
                r2l_avg = (stat_history[1]['r2l'] - stat_history[-1]['r2l'])/(stat_history[1]['ts'] - stat_history[-1]['ts'])
            else:
                l2r_avg = r2l_avg = 0.0

            # Trim the oldest points
            del(stat_history[stat_history_len+1:])

            uptime = seconds_to_human(time.time()-start_ts, decimal=0)
            l2r_t_h, l2r_t_u = bytes_to_human(stat_history[1]['l2r'])
            r2l_t_h, r2l_t_u = bytes_to_human(stat_history[1]['r2l'])
            l2r_a_h, l2r_a_u = bytes_to_human(l2r_avg)
            r2l_a_h, r2l_a_u = bytes_to_human(r2l_avg)

            _erase_line()
            print(f"{uptime} | In: {r2l_t_h:6.1f}{r2l_t_u:>2s} @ {r2l_a_h:6.1f}{r2l_a_u:>2s}/s | Out: {l2r_t_h:6.1f}{l2r_t_u:>2s} @ {l2r_a_h:6.1f}{l2r_a_u:>2s}/s", end="", flush=True)

def random_ips(network):
    # Network address
    net = ipaddress.ip_network(network)
    # Random host-part
    host_bytes = int(random.uniform(2, 2**(net.max_prefixlen-net.prefixlen)-4))&0xFFFFFFFE
    # Construct local/remote IP
    local_ip = net.network_address + host_bytes
    remote_ip = net.network_address + host_bytes + 1
    return local_ip.compressed, remote_ip.compressed

def main():
    if sys.platform != "linux":
        print("The 'ssm-tunnel' program only works on Linux at the moment!", file=sys.stderr)
        print("In other systems you are welcome to install it in VirtualBox or in a similar virtual environment running Linux.", file=sys.stderr)
        quit(1)

    ## Split command line args
    args = parse_args()

    logger = configure_logging(logger_name, args.log_level)

    tunnel = None
    try:
        if args.list:
            # --list
            InstanceResolver(args).print_list()
            quit(0)

        instance_id = InstanceResolver(args).resolve_instance(args.INSTANCE)
        if not instance_id:
            logger.warning("Could not resolve Instance ID for '%s'", args.INSTANCE)
            logger.warning("Perhaps the '%s' is not registered in SSM?", args.INSTANCE)
            quit(1)

        local_ip, remote_ip = random_ips(args.tunnel_cidr)
        tunnel = SsmTunnel(instance_id, profile=args.profile, region=args.region, logger_name=logger_name)
        tunnel.start(local_ip, remote_ip, args.routes or [], args.updown_script)

    except (botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError) as e:
        logger.error(e)
        quit(1)

    finally:
        if tunnel:
            tunnel.delete_tun()

if __name__ == "__main__":
    main()



# aws-ssm-tools packaging

import os
import sys
import pathlib

from setuptools import setup, find_packages
from setuptools.command.install import install

import ssm_tools

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

SCRIPTS = [
    'ssm-session',
    'ecs-session',
    'ssm-ssh',
    'ssm-tunnel',
]
VERSION = ssm_tools.__version__

requirements = HERE / "requirements.txt"
with requirements.open() as f:
    reqs = [ req.strip() for req in f.readlines() if req.strip() and not req.startswith('#') ]

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')
        if not tag:
            sys.exit("Env var $CIRCLE_TAG is not defined - are we running a CircleCI build?")

        if tag.startswith('v'): # If tag is v1.2.3 make it 1.2.3
            tag = tag[1:]

        if tag != VERSION:
            info = f"Git tag: {tag} does not match the version of this app: {VERSION}"
            sys.exit(info)

setup(
    name="aws-ssm-tools",
    version=VERSION,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            # All script entries must be in this format:
            # "ssm-session = ssm_tools.ssm_session_cli:main"
            f"{script} = ssm_tools.{script.replace('-','_')}_cli:main" for script in SCRIPTS
        ]
    },

    python_requires='>=3.6',

    install_requires=reqs,

    package_data={
        '': ['*.txt', '*.md', 'ssm-tunnel-updown.dns-example'],
    },

    author="Michael Ludvig",
    author_email="mludvig@logix.net.nz",
    description="Tools for AWS Systems Manager: "+" ".join(SCRIPTS),
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    keywords="aws ssm "+" ".join(SCRIPTS),
    url="https://github.com/mludvig/aws-ssm-tools",
    project_urls={
        "Bug Tracker": "https://github.com/mludvig/aws-ssm-tools/issues",
        "Documentation": "https://github.com/mludvig/aws-ssm-tools/blob/master/README.md",
        "Source Code": "https://github.com/mludvig/aws-ssm-tools",
    },

    classifiers=[
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: Networking',
    ],

    cmdclass={
        'verify': VerifyVersionCommand,
    },
)



#!/usr/bin/env python3

# Convenience wrapper around 'aws ssm start-session'
# can resolve instance id from Name tag, hostname, IP address, etc.
#
# See https://aws.nz/aws-utils/ssm-session for more info.
#
# Author: Michael Ludvig (https://aws.nz)

# The script can list available instances, resolve instance names,
# and host names, etc. In the end it executes 'aws' to actually
# start the session.

import os
import sys
import logging
import signal
import argparse
import botocore.exceptions

from .common import *
from .resolver import InstanceResolver

logger = logging.getLogger()

signal.signal(signal.SIGTSTP, signal.SIG_IGN)   # Ignore Ctrl-Z - pass it on to the shell

def parse_args(argv):
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)

    add_general_parameters(parser)

    group_instance = parser.add_argument_group('Instance Selection')
    group_instance.add_argument('INSTANCE', nargs='?', help='Instance ID, Name, Host name or IP address')
    group_instance.add_argument('--list', '-l', dest='list', action="store_true", help='List instances available for SSM Session')

    group_session = parser.add_argument_group('Session Parameters')
    group_session.add_argument('--user', '-u', '--sudo', dest='user', metavar="USER", help='SUDO to USER after opening the session. Can\'t be used together with --document-name / --parameters. (optional)')
    group_session.add_argument('--document-name', dest='document_name', help='Document to execute, e.g. AWS-StartInteractiveCommand (optional)')
    group_session.add_argument('--parameters', dest='parameters', help='Parameters for the --document-name, e.g. \'command=["sudo -i -u ec2-user"]\' (optional)')

    parser.description = 'Start SSM Shell Session to a given instance'
    parser.epilog = f'''
IMPORTANT: instances must be registered in AWS Systems Manager (SSM)
before you can start a shell session! Instances not registered in SSM
will not be recognised by {parser.prog} nor show up in --list output.

Visit https://aws.nz/aws-utils/ssm-session for more info and usage examples.

Author: Michael Ludvig
'''

    # Parse supplied arguments
    args, extras = parser.parse_known_args(argv)

    # If --version do it now and exit
    if args.show_version:
        show_version(args)

    # Require exactly one of INSTANCE or --list
    if bool(args.INSTANCE) + bool(args.list) != 1:
        parser.error("Specify either INSTANCE or --list")

    if args.parameters and not args.document_name:
        parser.error("--parameters can only be used together with --document-name")

    if args.user and args.document_name:
            parser.error("--user can't used used together with --document-name")

    return args, extras

def start_session(instance_id, args):
    aws_args = ""
    if args.profile:
        aws_args += f"--profile {args.profile} "
    if args.region:
        aws_args += f"--region {args.region} "

    ssm_args = ""
    if args.user:
        # Fake --document-name / --parameters for --user
        ssm_args += f"--document-name AWS-StartInteractiveCommand --parameters 'command=[\"sudo -i -u {args.user}\"]'"
    else:
        # Or use the provided values
        if args.document_name:
            ssm_args += f"--document-name {args.document_name} "
        if args.parameters:
            ssm_args += f"--parameters '{args.parameters}' "

    command = f'aws {aws_args} ssm start-session --target {instance_id} {ssm_args}'
    logger.info("Running: %s", command)
    os.system(command)

def main():
    ## Split command line to main args and optional command to run
    args, extras = parse_args(sys.argv[1:])

    global logger
    logger = configure_logging("ssm-session", args.log_level)

    try:
        instance = None
        if args.list:
            InstanceResolver(args).print_list()
            quit(0)

        instance_id = InstanceResolver(args).resolve_instance(args.INSTANCE)

        if not instance_id:
            logger.warning("Could not resolve Instance ID for '%s'", args.INSTANCE)
            logger.warning("Perhaps the '%s' is not registered in SSM?", args.INSTANCE)
            quit(1)

        start_session(instance_id, args)

    except (botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError) as e:
        logger.error(e)
        quit(1)

if __name__ == "__main__":
    main()
